// Standalone AVR ISP programmer
// August 2011 by Limor Fried / Ladyada / Adafruit
// Jan 2011 by Bill Westfield ("WestfW")
//
// Tweaks by Mike Tsao
// Further tweaks by Thomas Laubach, now works with current SPI library (April 2016)

// This sketch allows an Arduino to program a flash program
// into any AVR if you can fit the HEX file into program memory.
// No computer is necessary. Two LEDs for status notification.
// Press button to program a new chip. Piezo beeper for error/success
// This is ideal for very fast mass-programming of chips!
//
// It is based on AVRISP
//
// Using the following pins:
// 10: slave select
// 11: MOSI
// 12: MISO
// 13: SCK
//  9: 8 MHz clock output - connect this to the XTAL1 pin of the AVR
//     if you want to program a chip that requires a crystal without
//     soldering a crystal in
// ----------------------------------------------------------------------

/*
atmega1284p hat ein 256 word window (Seite 297 Data sheet, 25.7.2)
ein Wort entspricht 2 Bytes, also hat der atmeg1284p eine flash page size = 256/2 = 128 byte,
identisch zum atmega328p (verbaut im Arduino Uno), siehe auch http://www.mikrocontroller.net/articles/AVR_Bootloader_in_C_-_eine_einfache_Anleitung#Schritt_4_-_Flashen_und_Ausprobieren_des_Bootloaders_2
*/

#include "optiLoader.h"
#include "SPI.h"
#include <SD.h>
#include <avr/pgmspace.h>
//#include "optiboot.h"


// Global Variables
int pmode = 0;
byte pageBuffer[128];                  /* One page of flash */

/*
/*
 * Pins to target
 */
 /*
// ARDUINO UNO

#define SCK 13
#define MISO 12
#define MOSI 11
*/

// SPI Slaves
#define SS_SDCARD     53 // = Slave Select of SPI device "SD card reader"
#define RESET 10 // = Slave Select of SPI device "Satshakit 128"


// Programmer: ARDUINO MEGA 2560:
#define SCK 52
#define MISO 50
#define MOSI 51

/*
// SATSHAKIT 128
#define SCK 7
#define MISO 6
#define MOSI 5
*/



#define CLOCK 9     // self-generate 8mhz clock - handy!
#define BUTTON A1

SPISettings settingsA(16000000/128, MSBFIRST, SPI_MODE0); // ADDED BY THOMAS

// SD card preparations
File myFile;
File root;
uint32_t sdimage_size; // Size of game file

void setup() 
{
  Serial.begin(57600);  // Initialize serial for status msgs
  Serial.println("\nAdaBootLoader Bootstrap programmer "
                 "(originally OptiLoader Bill Westfield (WestfW))");
  Serial.println("Modified by Mike Tsao, further modified by Thomas Laubach\n");

  pinMode(BUTTON, INPUT);     // button for next programming
  digitalWrite(BUTTON, HIGH); // pullup

  // Initialize SPI devices
  pinMode(RESET, OUTPUT); // Target Satshakit 128
  pinMode(SS_SDCARD, OUTPUT);
  delay(100);
  
  pinMode(CLOCK, OUTPUT);
  // set up high freq PWM on pin 9 (timer 1)
  // 50% duty cycle -> 8 MHz
  OCR1A = 0;
  ICR1 = 1;
  // OC1A output, fast PWM
  TCCR1A = _BV(WGM11) | _BV(COM1A1);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10); // no clock prescale
  
  
  // --- INITIALIZE SD-CARD READER ---

  digitalWrite(RESET, HIGH); // Deactivate SPI device Satshakit 128
  digitalWrite(SS_SDCARD, LOW); // Activate SPI device SD card reader
  Serial.print("Initializing SD card...");
  // On the Ethernet Shield, CS is pin 4. It's set as an output by default.
  // Note that even if it's not used as the CS pin, the hardware SS pin
  // (10 on most Arduino boards, 53 on the Mega) must be left as an output
  // or the SD library functions will not work.

  if (!SD.begin(SS_SDCARD))
  {
    Serial.println("initialization failed!");
    return;
  }
  Serial.println("initialization done.");
  
  root = SD.open("/");
  printDirectory(root, 0);
    
  // --- LET THE USER SELECT A GAME FROM SD CARD: STUB ---

  File file = selectGameFromSDCard();
  Serial.print("Will burn following file onto the Satshakit 128: "); Serial.print(file.name()); Serial.print("\n");
  digitalWrite(SS_SDCARD, HIGH); // Deactivate SD card reader
  SPI.endTransaction();
  delay(300);
  
  // -- OBTAIN DATA STRUCTURE WITH INFORMATION ON TARGET --
  
  uint16_t signature;
  image_t *targetimage;
  signature = 0x9705;
  targetimage = findImage(signature);
  
  // --- ERASE THE FLASH MEMORY OF THE TARGET ATMEGA1284P ---

  SPI.beginTransaction(settingsA);
  target_poweron();
  //end_pmode();
  //start_pmode();
  digitalWrite(RESET, LOW); // Activate SPI device Satshakit 128
  eraseChip();
  delay(100);
  
  // --- PROGRAM AND VERIFY THE FUSES OF THE TARGET ATMEGA1284P ---
  
  // Get fuses ready to program
  if (!programFuses(targetimage->image_progfuses)) 
  {
    error("\nProgramming Fuses fail");
    return;
  }

  if (!verifyFuses(targetimage->image_progfuses, targetimage->fusemask)) 
  {
    error("\nFailed to verify fuses");
    return;
  }
  
  // --- READ GAME PAGE PER PAGE FROM SD-CARD, BURN IT PAGE PER PAGE ONTO TARGET
  
  end_pmode();
  start_pmode();

  // Necessary variables for writing to the target device's Flash Memory
  const byte* hex = targetimage->hexcode; // is byte hexcode[3072],  Intel hex format image
  boolean as_hex = pgm_read_byte(hex) == ':';
  uint32_t pageaddr = 0;
  uint8_t pagesize = pgm_read_byte(&targetimage->image_pagesize);
  //uint16_t chipsize = pgm_read_word(&targetimage->chipsize);
  uint32_t chipsize = 131072;
  // Necessary variables for reading from the SD card
  sdimage_size = file.size(); 
  /*
  Serial.print("\npageaddr, pagesize chipsize: "); 
  Serial.print(pageaddr); Serial.print(" "); Serial.print(pagesize); 
  Serial.print(", "); Serial.print(chipsize, DEC);
  Serial.print("\nErste Page aus hex: \n");
  for (int i = 0; i < 256; i++) 
  {
    Serial.print(targetimage->hexcode[i]); Serial.print(", ");
  }
  Serial.println("....184 in HEX = "); Serial.print(184, HEX);
  Serial.println("...0x10, 0x00 in DEC = "); Serial.print(0x1000, DEC); 
  while (true) {}
  */
  long pageNum = 1; // Number of pages occupied by the program
  long numPages = floor(sdimage_size / 128);
  byte pRemainder = sdimage_size % 128; // Number of bytes on the last page occupied by the program
  byte p = 0; // Counter within pBuffer
  long byteCntFile = 0; // Counts how many bytes from the file have already been read
  byte byteInPage = 0;

  Serial.print("\npageaddr: "); Serial.print(" "); Serial.print(pageaddr); Serial.print("\n");
  while (pageaddr < sdimage_size)
  {
    end_pmode(); // Set inactive SPI device "Satshakit 128"
    digitalWrite(SS_SDCARD, LOW); // Set active SPI device "SD-CARD READER" 
    
    // We are on the last page in the file
    if (pageNum == numPages)
    {
      for (uint8_t i = 0; i < pagesize; i++)
      {
        // Fill the page buffer with what remains in the file,
        // fill the remaining part with 0xFF
        if (i <= pRemainder) { pageBuffer[i] = file.read(); }
        else { pageBuffer[i] = 0xFF; }
      }
    } // on last page in the file
    else // on any page in the file but the last
    {
      // Empty the page buffer by filling it with 0xFF's
      for (uint8_t i = 0; i < pagesize; i++) pageBuffer[i] = 0xFF;
      // Fill the whole page with the next 128 bytes from the file
      for (uint8_t i = 0; i < pagesize; i++)
      {
        pageBuffer[i] = file.read(); // Read next byte from file into buffer
        Serial.print(pageBuffer[i], HEX); Serial.print(" ");
      }
    } // on any page in the file but the last

    // Verify whether the page buffer is empty
    boolean blankpage = true;
    for (uint8_t i = 0; i < pagesize; i++) { if (pageBuffer[i] != 0xFF) blankpage = false; }
    
    // The page buffer has been filled with the next 128 bytes from the file:
    // Burn the page onto the ATmega1284p
    if (!blankpage)  
    {
      start_pmode(); // Set active SPI device "Satshakit 128"
      digitalWrite(SS_SDCARD, HIGH); // Set inactive SPI device "SD-CARD READER"
      /*
      if (!flashPage(pageBuffer, pageaddr, pagesize)) 
      {
        error("Flash programming failed");
        return;
      }
      */
    }

    // Move to next page
    pageNum++;
    pageaddr = pageaddr + pagesize;
    Serial.print("\npageaddr: "); Serial.print(" "); Serial.print(pageaddr); Serial.print("\n");

    byteCntFile++;
    byteInPage++;  
  } // while

while (true) {}

     
  // Set fuses to 'final' state
  if (!programFuses(targetimage->image_normfuses)) 
  {
    error("Programming Fuses fail");
    return;
  }

  end_pmode();
  start_pmode();
  
  if (!verifyFuses(targetimage->image_normfuses, targetimage->fusemask) ) 
  {
    error("Failed to verify fuses");
  } 
  else 
  {
    Serial.println("Fuses correct");
  }
  target_poweroff();
}


File selectGameFromSDCard()
{
  File file;
  file = SD.open("BLINK");
  return file;
}



void loop(void) 
{
}

void error(const char *string) 
{
  while (!digitalRead(BUTTON))
    ;
  Serial.println(string);
  target_poweroff();
  while (digitalRead(BUTTON)) 
  {
    digitalWrite(LED_ERR, HIGH);
    digitalWrite(LED_PROGMODE, LOW);
    //tone(PIEZOPIN, 622, 500);
    delay(500);
    digitalWrite(LED_ERR, LOW);
    digitalWrite(LED_PROGMODE, HIGH);
    //tone(PIEZOPIN, 460, 500);
    delay(500);
  }
  while (!digitalRead(BUTTON))
    ;
  digitalWrite(LED_ERR, LOW);
  digitalWrite(LED_PROGMODE, LOW);
}

void start_pmode() 
{
  //pinMode(13, INPUT); // restore to default
  pinMode(SCK, INPUT); // restore to default
  //SPI.begin(); // REMOVED BY THOMAS
  //SPI.setClockDivider(SPI_CLOCK_DIV128); // REMOVED BY THOMAS
  SPI.beginTransaction(settingsA); // ADDED BY THOMAS

  debug("...spi_init done");
  // following delays may not work on all targets...
  pinMode(RESET, OUTPUT);
  digitalWrite(RESET, HIGH);
  pinMode(SCK, OUTPUT);
  digitalWrite(SCK, LOW);
  delay(100);
  digitalWrite(RESET, LOW);
  delay(100);
  pinMode(MISO, INPUT);
  pinMode(MOSI, OUTPUT);
  debug("...spi_transaction");
  Serial.println("\nIn start_pmode: spi_transaction");
  spi_transaction(0xAC, 0x53, 0x00, 0x00);
  debug("...Done");
  Serial.println("did spi_transaction");
  pmode = 1;
}

void end_pmode() 
{
  SPCR = 0;                             /* reset SPI */
  digitalWrite(MISO, 0);                /* Make sure pullups are off too */
  pinMode(MISO, INPUT);
  digitalWrite(MOSI, 0);
  pinMode(MOSI, INPUT);
  digitalWrite(SCK, 0);
  pinMode(SCK, INPUT);
  digitalWrite(RESET, 0);
  pinMode(RESET, INPUT);
  pmode = 0;
}

/*
 * target_poweron
 * begin programming
 */
boolean target_poweron() 
{
  //pinMode(LED_PROGMODE, OUTPUT);
  //digitalWrite(LED_PROGMODE, HIGH);
  digitalWrite(RESET, LOW);  // reset it right away.
  pinMode(RESET, OUTPUT);
  delay(100);
  Serial.print("Starting Program Mode");
  start_pmode();
  Serial.println(" [OK]");
  return true;
}

boolean target_poweroff() 
{
  end_pmode();
  //digitalWrite(LED_PROGMODE, LOW);
  return true;
}

bool isHexFile(char* filename) 
{
    int8_t len = strlen(filename);
    bool result;
    if (  strstr(strlwr(filename + (len - 4)), ".hex")
          || strstr(strlwr(filename + (len - 4)), ".HEX")
       ) 
    {
      result = true;
    } 
    else 
    {
      result = false;
    }
}



void printDirectory(File dir, int numTabs)
{
  byte b;
  
  while (true) 
  {
    File entry =  dir.openNextFile();
    if (! entry) 
    {
      // no more files
      break;
    }
    
    if (!entry.isDirectory()) 
    {
      if ( isHexFile(entry.name()))
      {
        Serial.print(entry.name());
        Serial.print("\n");

        sdimage_size = entry.size();
        Serial.println(sdimage_size);
        //break;
      }
    }
    entry.close();
  } // while
}


